@@ -23,6 +23,7 @@ import locale
 import logging
 import logging.config
 import multiprocessing
+import operator
 import os
 import os.path as osp
 from os.path import abspath, basename, dirname, expanduser, exists, isfile, join, splitext
@@ -663,7 +664,27 @@ def profile_runs(funcname, modulefile, nruns=5):
     stats.print_stats()
 
 
-def get_local_tmp_dir():
+def get_platform_home_dir():
+    plat = platform.system()
+    if plat == 'Windows':
+        return os.getenv('USERPROFILE')
+    elif plat == 'Darwin':
+        return osp.expanduser('~/')
+    elif plat == 'Linux':
+        return osp.expanduser('~/')
+    raise NotImplementedError(f'unsupported platform: {plat}')
+
+
+def get_platform_appdata_dir(winroam=True):
+    plat = platform.system()
+    if plat == 'Windows':
+        return os.getenv('APPDATA' if winroam else 'LOCALAPPDATA')
+    elif plat == 'Darwin':
+        return osp.expanduser('~/Library/Application Support')
+    raise NotImplementedError(f'unsupported platform: {plat}')
+
+
+def get_platform_tmp_dir():
     plat = platform.system()
     if plat == 'Windows':
         return join(os.getenv('LOCALAPPDATA'), 'Temp')
@@ -791,7 +812,7 @@ class RerunLock:
     def __init__(self, name, folder=None, logger=_logger):
         os.makedirs(folder, exist_ok=True)
         filename = f'lock_{name}.json' if name else 'lock_{}.json'.format(next(tempfile._get_candidate_names()))
-        self.lockFile = osp.join(folder, filename) if folder else join(get_local_tmp_dir(), filename)
+        self.lockFile = osp.join(folder, filename) if folder else join(get_platform_tmp_dir(), filename)
         self.logger = logger
         common_sigs = [
             signal.SIGABRT,
@@ -967,48 +988,48 @@ def remove_from_os_paths(bindir, usesyspath=True):
     os.environ[path_var] = os.environ[path_var].replace(bindir, '')
 
 
-def run_cmd(cmd, cwd='.', logger=None, check=True):
-    if logger:
-        logger.debug(' '.join(cmd))
-    else:
-        print(' '.join(cmd))
-    proc = None
+def run_cmd(cmd, cwd=None, logger=None, check=True):
+    local_debug = logger.debug if logger else print
+    local_info = logger.info if logger else print
+    local_error = logger.error if logger else print
+    local_debug(f"""\
+{' '.join(cmd)}
+cwd: {osp.abspath(cwd) if cwd else os.getcwd()}
+""")
     try:
         proc = subprocess.run(cmd, check=check, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
         log = proc.stdout.decode(TXT_CODEC)
-        if logger:
-            logger.debug(log)
-        else:
-            print(log)
+        local_debug(log)
     except subprocess.CalledProcessError as e:
         log = f'stdout: {e.stdout.decode(TXT_CODEC)}'
         err = f'stderr: {e.stderr.decode(TXT_CODEC)}'
-        if logger:
-            logger.info(log)
-            logger.error(err)
-        else:
-            print(log)
-            print(err)
+        local_info(log)
+        local_error(err)
         raise e
     except Exception as e:
-        if logger:
-            logger.error(e)
+        # no need to have header, exception has it all
+        local_error(e)
         raise e
     return proc
 
 
-def run_daemon(cmd, cwd='.', logger=None):
+def run_daemon(cmd, cwd=None, logger=None, check=True):
+    local_debug = logger.debug if logger else print
+    local_info = logger.info if logger else print
+    local_error = logger.error if logger else print
+    local_debug(f"""run in ground:
+{" ".join(cmd)}
+cwd: {osp.abspath(cwd) if cwd else os.getcwd()}
+""")
     try:
         proc = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
         # won't be able to retrieve log from background
     except subprocess.CalledProcessError as e:
-        if logger:
-            logger.info(f'stdout: {e.stdout.decode(TXT_CODEC)}')
-            logger.error(f'stderr: {e.stderr.decode(TXT_CODEC)}')
+        local_info(f'stdout: {e.stdout.decode(TXT_CODEC)}')
+        local_error(f'stderr: {e.stderr.decode(TXT_CODEC)}')
         raise e
     except Exception as e:
-        if logger:
-            logger.error(e)
+        local_error(e)
         raise e
     return proc
 
@@ -1358,7 +1379,7 @@ def compare_textfiles(file1, file2, showdiff=False, contextonly=True, logger=Non
         if showdiff:
             diff_func = difflib.context_diff if contextonly else difflib.Differ().compare
             diff = diff_func(lines1, lines2)
-            lazy_logging(f'*** {file1} vs. {file2} ***')
+            lazy_logging(f'***\n{file1} vs.\n{file2}\n***')
             lazy_logging(''.join(diff), logger)
     return lines1 == lines2
 
@@ -1499,6 +1520,24 @@ def open_in_editor(path):
     run_cmd(cmd, check=check)
 
 
+def flatten_nested_lists(mylist):
+    return functools.reduce(operator.iconcat, mylist, [])
+
+
+def show_results(succeeded, detail, advice, dryrun=False):
+    banner = '** DRYRUN **' if dryrun else '*** SUCCEEDED ***' if succeeded else '* FAILED *'
+    advice_title = '' if not advice else 'Next:' if succeeded else 'Advice:'
+    report = f"""
+{banner}
+
+{detail}
+
+{advice_title}
+{advice}
+"""
+    print(report)
+
+
 def _test():
     pass
 
